/****************************************************************************
* - DIY remote control XBee support file
*
****************************************************************************/
#define DEBUG
//#define DEBUG_OUT
//#define DEBUG_VERBOSE

#include "BapLib.h"
#include "diyxbee.h"
#include "hsio.h"
#include <stdarg.h>

#ifdef __cplusplus
extern "C" {
#endif



DIYSTATE g_diystate;

bool g_fDebugOutput;

// Some forward definitions
#ifdef DEBUG
extern void DebugMemoryDump(const byte* , int, int);
static bool s_fDisplayedTimeout = false;
#endif




//=============================================================================
// Xbee stuff
//=============================================================================
void InitXBee(u8 bpinIn, u8 bpinOut, u8 bpinRTS, word wBaudMode)
{
	// Save away the io pins that we are using.
	g_diystate.bpinIn = bpinIn;
	g_diystate.bpinOut = bpinOut;
	g_diystate.bpinRTS = bpinRTS;
	g_diystate.wBaudMode = wBaudMode;
	g_fDebugOutput = true;
	if (g_diystate.bpinIn == _S_HSERIAL_) {
		DBGPrintf("Before EnableHSerial\r");
		enablehserial();
		DBGPrintf("Before SetHSerial\r");
		sethserial(g_diystate.wBaudMode, 0,0,0);
		DBGPrintf("After SetHSerial\r");
	}

	// Our IOPrintf can handle both standard as well as hserial so can cleanup!
	if (g_diystate.bpinRTS != (u8)-1) {
		XBeePrintf("+++");
		WaitForXBeeTransmitComplete();
		pause(20);
		XBeePrintf("ATD6 1\rATCN\r");
		output(g_diystate.bpinRTS);						// Make sure the pin is set for input.
		low (g_diystate.bpinRTS);						// make sure RTS is output and input starts off enabled...
	}
	

	// for Xbee with no flow control, maybe nothing to do here yet...
	g_diystate.fTransReadyRecvd = false;
	g_diystate.fPacketValid = false; 
	g_diystate.bPacketNum = 0;
}


//=============================================================================
// byte ReadFromXBee - Read in a buffer of bytes.  We will pass in a timeout
//			value that if we dont receive a character in that amount of time 
//			something is wrong.  
//=============================================================================

#ifdef DEBUG_VERBOSE
byte ReadFromXBee(byte *pb, byte cb, ulong Timeout, word wEOL)
{
	byte cbRead;
	byte* pbT = pb;

	if (g_fDebugOutput) {
		DBGPrintf("Read From XBee: cbRead: %d\n\r", cbRead);
		if (cbRead > 0) 
			DebugMemoryDump(pbT, 0, cbRead);
	}

	return cbRead;
}

#else

inline byte ReadFromXBee(byte *pb, byte cb, ulong Timeout, word wEOL) __attribute__((always_inline));

byte ReadFromXBee(byte *pb, byte cb, ulong Timeout, word wEOL)
{
	if (g_diystate.bpinIn == _S_HSERIAL_)
		return hserin(Timeout, pb, cb, wEOL);

	return TASerin(g_diystate.bpinIn, g_diystate.bpinRTS, g_diystate.wBaudMode, (char*)pb, cb, Timeout, wEOL);
}


#endif



//=============================================================================
// void WriteToXBee - output a buffer to the XBEE.  This is used to abstract
//			away the actual function needed to talk to the XBEE
//=============================================================================
inline void WriteToXBee(byte *pb, byte cb) __attribute__((always_inline));
void WriteToXBee(byte *pb, byte cb)
{
	if (g_diystate.bpinIn == _S_HSERIAL_)
		hserout(pb, cb);
	else
		TASerout(g_diystate.bpinOut, g_diystate.wBaudMode, (char*)pb, cb);
}

//=============================================================================
// void XBeePrintf - output a buffer to the XBEE.  This is used to abstract
//			away the actual function needed to talk to the XBEE
//=============================================================================
inline int XBeePrintf(const char *format, ...) __attribute__((always_inline));
int XBeePrintf(const char *format, ...)
{
	va_list ap;
	va_start(ap, format);
	return IOVPrintf(g_diystate.bpinOut, g_diystate.wBaudMode, format, ap);
}


//==============================================================================
// [SendXBeePacket] - Simple helper function to send the 4 byte packet header
//	 plus the extra data if any
// 	 gosub SendXBeePacket[bPacketType, bSeq, cbExtra, pExtra]
//==============================================================================
void SendXBeePacket(byte bPHType, byte bSeq, byte cbExtra, byte *pbExtra)
{
	// Tell system to now output to the xbee
	byte abPH[4];
	byte *pbT;
	
	abPH[0] = bPHType;
	abPH[2] = bSeq;
	abPH[3] = cbExtra;
	
	abPH[1] = bSeq + cbExtra;		// start the checksum;
	
	pbT = pbExtra;
	
	// loop through the extra bytes in the exta to build the checksum;
	while (cbExtra--)
		abPH[1] += *pbT++;				// add each byte to the checksum

	// Now output the buffers.  Need to integrate more
	WriteToXBee(abPH,4);	// Set the destination address
	if (abPH[3])
		WriteToXBee(pbExtra, abPH[3]);

#ifdef DEBUG_OUT
	// We moved dump before the serout as hserout will cause lots of interrupts which will screw up our serial output...
	// Moved after as we want the other side to get it as quick as possible...
	DBGPrintf("SDP: %2x %2x %2x %2x :", abPH[0], abPH[1], abPH[2], abPH[3]);
	
#ifdef DEBUG_VERBOSE		// Only ouput whole thing if verbose...
	if (abPH[3]) {
		byte i;
		for (i = 0; i < abPH[3], i++) {
			DBGPrintf(" %2x", pbExtra++);
		}
	}
#endif	
	DBGPrintf("\r");
		
#endif

}


//==============================================================================
// [SendXBeePacketAPIMode] - Simple helper function will send a packet while
//		we are in API mode.  We will first start off with the data having the
//		old format, and then once we have everything working on both ends will
//		streamline here as no use in duplicating checksums and the like...
// 	 SendXBeePacket(bPacketType, bSeq, cbExtra, pExtra)
//		Note: we are not using the escaped char mode.
//==============================================================================
void SendXBeePacketAPIMode(word wDL, byte bPHType, byte bSeq, byte cbExtra, byte *pbExtra)
{
#ifdef LATER
	byte	bChkSum ;
	byte	*pbT;
	short	i;

	// In API mode the format of message is for 16 bit address 
	// 7E <len MSB> <len LSB> 01 <frame id> <address MSB> <Address lsb> <options> [Our data] <chksum>

	// Tell system to now output to the xbee
	rprintfInit(WIRELESS_ACTIVATE);
	rprintfChar (0x7e);			// command start
	rprintfChar (0x00);			// assume our data will not exceed 250 characters so high byte of size will be zero
	rprintfChar	(cbExtra+4+5);	// 5 bytes for data packet bytes + 4 bytes of our own data + however many extra bytes...
	rprintfChar (0x01);			// API = 1 - Send data by 16 bit address
	rprintfChar (bSeq);			// Frame id - might as well use the sequence number we passed in


	rprintfChar (wDL >> 8);		// Destination high byte
	rprintfChar (wDL & 0xff);	// Destination low byte

	rprintfChar (0);			// no options


	bChkSum = bSeq + cbExtra;

	if (cbExtra) {
		pbT = pbExtra;
		for (i=0; i < cbExtra; i++) {
			bChkSum += *pbT++;		// walk through the extra bytes
		}
	}
#ifdef DEBUG_VERBOSE
	if (g_fDebugOutput) {
		DBGPrintf("SendXBeePacket:%x %x %x %x\n", bPHType,bChkSum, bSeq, cbExtra);
		if (cbExtra > 0) 
			DebugMemoryDump(pbExtra, 0, cbExtra);
	}
#endif //DEBUG
	// Now lets output our actual data for the packet...
	rprintfChar(bPHType);
	rprintfChar(bChkSum);
	rprintfChar(bSeq);
	rprintfChar(cbExtra);

	for (i=0; i < cbExtra; i++) {
		rprintfChar(*pbExtra++);
	}

	// Last but not least we need to output the XBEE checksum for this whole packet...
	// Add on the bytes that were not used in our own check sum + the 5 bytes of the API msg data
	bChkSum += (bChkSum + bPHType) + (01 + bSeq + (wDL >> 8) + (wDL & 0xff) + 0);
	rprintfChar(0xff-bChkSum);
#endif
}

//==============================================================================
// [SetXbeeDL] - Set the XBee DL to the specified word that is passed
// BUGBUG: sample function, need to clean up.
//==============================================================================
void SetXBeeDL (unsigned short wNewDL)
{
#ifdef DEBUG_VERBOSE
	IOPrintf(_S_OUT_, I9600, "Set XBEE DL: %x\n\r", wNewDL);
#endif

	pause(20);
	XBeePrintf("+++");
	WaitForXBeeTransmitComplete();

	pause(20);
	XBeePrintf("ATDL %x\rATCN\r", wNewDL);
	pause(10);
}


//==============================================================================
// [GetXBeeHVal] - Set the XBee DL or MY or??? Simply pass the two characters
// 			that were passed in to the XBEE
//==============================================================================
word GetXBeeHVal (char c1, char c2)
{
	word wDL;
	byte bTemp[10];						// temporary buffer
	byte cbRead;
	byte i;	


	ClearXBeeInputBuffer();				// Clear our input buffer
	wDL = (word)-1;						// an error condition
	pause(20);							// wait 20ms for quard time to enter into command mode
	XBeePrintf("+++");

	WaitForXBeeTransmitComplete();
	pause(20);						// wait 20ms for quard time to enter into command mode

	XBeePrintf("AT%c%c\rATCN\r", c1, c2);	// Ok I output the new DL as well as the ATCN to exit command mode

	// Now lets try to read some stuff in
	cbRead = ReadFromXBee(bTemp, sizeof(bTemp), 10000, 13);	// This should be OK<cr>

	cbRead = ReadFromXBee(bTemp, sizeof(bTemp), 10000, 13);	// This shoould be our DL<cr>

	if ((cbRead > 1) && (bTemp[cbRead-1] == 13)) {
		// We need to convrt the number from a hex string to a number...
		wDL = 0;	// start off with 0
		cbRead--;	// subtract off the cr...
		for (i = 0; i < cbRead; i++) {
			if ((bTemp[i] >= '0') && (bTemp[i] <= '9'))
				wDL = (wDL<<4) + (bTemp[i]-'0');
			else if ((bTemp[i] >= 'A') && (bTemp[i] <= 'F'))
				wDL = (wDL<<4) + (bTemp[i]-'A' + 10);
			else if ((bTemp[i] >= 'a') && (bTemp[i] <= 'f'))
				wDL = (wDL<<4) + (bTemp[i]-'a' + 10);
		}
	}
	cbRead = ReadFromXBee(bTemp, sizeof(bTemp), 10000, 13);	// This should be OK<cr>
#ifdef DEBUG
	if (g_fDebugOutput) {
		DBGPrintf("GetXBEEHVAL: %c%c=%x\n", c1, c2, wDL);
	}
#endif //DEBUG

	return wDL;
}


	
//==============================================================================
// [ClearXBeeInputBuffer] - This simple helper function will clear out the input
//						buffer from the XBEE
//==============================================================================
void ClearXBeeInputBuffer(void)
{
	byte b[1];
	if (g_diystate.bpinIn == _S_HSERIAL_)
		hserstat(HSERSTAT_CLEAR_INPUT_BUFFER);

#ifdef DEBUG
	bool fBefore = g_fDebugOutput;
	g_fDebugOutput = false;
#endif	
		
	while (ReadFromXBee(b, 1, 5000, 0xffff))
		;	// just loop as long as we receive something...
#ifdef DEBUG
	g_fDebugOutput = fBefore;
#endif	
}



//==============================================================================
// [WaitForXBeeTransmitComplete] - This simple helper function will loop waiting
//						for the uart to say it is done.
//==============================================================================
void WaitForXBeeTransmitComplete(void)
{
	// Only need to do something if we are a hardware serial port
	if (g_diystate.bpinIn == _S_HSERIAL_)
	while(!hserstat(HSERSTAT_OUTPUT_EMPTY))
		;
}

//==============================================================================
// [DebugMemoryDump] - striped down version of rprintfMemoryDump
//==============================================================================
#ifdef DEBUG
void DebugMemoryDump(const byte* data, int off, int len)
{
    int x;
    int c;
    int line;
    const byte * b = data;

    for(line = 0; line < ((len % 16 != 0) ? (len / 16) + 1 : (len / 16)); line++)  {
    	int line16 = line * 16;
    	DBGPrintf("%4x| ", line16);

        // put hex values
        for(x = 0; x < 16; x++) {
            if(x + line16 < len) {
                c = b[off + x + line16];
				DBGPrintf("%2x ", c);
            }
            else
				DBGPrintf("   ");
        }
		DBGPrintf("| ");

        // put ascii values
        for(x = 0; x < 16; x++) {
            if(x + line16 < len) {
                c = b[off + x + line16];
                DBGPrintf("%c",((c > 0x1f) && (c < 0x7f))? c : '.');
             }
             else
             	DBGPrintf(" ");
        }
        DBGPrintf("\n\r");
    }
}

#endif



__inline__ u32 get_sp(void) __attribute__ ((always_inline));
__inline__ u32 get_sp(void)
{
      u32 sp_val;
      asm("mov.l er7, %0":"=g"(sp_val));
      return(sp_val);
}


//==============================================================================
// [ReceiveXBeePacket] - This function will try to receive a packet of information
// 		from the remote control over XBee.
//
// the data in a standard packet is arranged in the following byte order:
//	0 - Buttons High
//	1 - Buttons Low
// 	2 - Right Joystick L/R
//	3 - Right Joystick U/D
//	4 - Left Joystick L/R
//	5 - Left Joystick U/D
// 	6 - Right Slider
//	7 - Left Slider
//==============================================================================
bool ReceiveXBeePacket(PDIYPACKET pdiyp)
{
	byte cbRead;
	byte bcsIn;
	byte *pb;
	int	i;
	XPACKETHEADER XBeePacketHeader;						// A packet header to use
	word wNewDL;
	ulong ulCurrentTime;
	ulong ulTimeDiffMS;
	bool _fPacketValidPrev = g_diystate.fPacketValid;		// Save away the previous state as this is the state if no new data...
	bool _fNewPacketAvail = false;
	
	g_diystate.fPacketValid = false;
	g_diystate.fPacketTimeOut = false;
	g_diystate.fPacketForced = false;

	//	We will first see if we have a packet header waiting for us.
	// will also return true if there is no input queue for the device
		
	if ((g_diystate.bpinIn != _S_HSERIAL_) || hserstat(HSERSTAT_INPUT_AVAILABLE)){
		// The XBEE has sent us some data so try to get a packet header
#ifdef DEBUG
		s_fDisplayedTimeout = false;		// say that we got something so we can display empty again...
#endif

		cbRead = ReadFromXBee((byte*)&XBeePacketHeader, sizeof(XPACKETHEADER), (ulong)10000, (word)0xffff);
#ifdef DEBUG_OUT
		// We moved dump before the serout as hserout will cause lots of interrupts which will screw up our serial output...
		// Moved after as we want the other side to get it as quick as possible...
		DBGPrintf("RPH(%d): %2x %2x %2x %2x :", cbRead, XBeePacketHeader.bPacketType, XBeePacketHeader.bChksum, XBeePacketHeader.bSeqNum, XBeePacketHeader.cbExtra);
#endif
		if (cbRead == sizeof(XPACKETHEADER)) {
			if ((XBeePacketHeader.bPacketType == XBEE_TRANS_DATA) && (XBeePacketHeader.cbExtra == 8)) {
				//-----------------------------------------------------------------------------
				// [XBEE_TRANS_DATA]
				//-----------------------------------------------------------------------------
				// process first as higher number of these come in...
				cbRead = ReadFromXBee((byte*)pdiyp, (byte)sizeof(DIYPACKET), (ulong)25000, (word)0xffff);
				if (cbRead == sizeof(DIYPACKET)) {

					g_diystate.fReqDataPacketSent = 0;	// if we received a packet reset so we will request data again...
					// validate the header information...
					// Ok Lets try to validate the checksum.
					bcsIn = XBeePacketHeader.bSeqNum + XBeePacketHeader.cbExtra;
					pb = (byte*)pdiyp;
					for (i = 0; i < XBeePacketHeader.cbExtra; i++)
						bcsIn += *pb++;	// walk through the bytes and add to checksum...

					if (bcsIn == XBeePacketHeader.bChksum) {
						// BUGBUG:: the order of the two bytes making up the wButtons is reversed so swap them here
						word wT = pdiyp->wButtons;
						pdiyp->wButtons = (wT >> 8) | ((wT&0xff) << 8);
						
						g_diystate.fPacketValid = true;	// data is valid
						g_diystate.fPacketForced = g_diystate.fReqDataForced;	// Was the last request forced???
						g_diystate.fReqDataForced = 0;				// clear that state now
						g_diystate.ulLastPacket = GetCurrentTime();
						return true;	//  		// get out quick!
					} else {
						// the checksum and data not right lets clear our input buffer out...	
						ClearXBeeInputBuffer();
					}
				}
			}
			else if ((XBeePacketHeader.bPacketType == XBEE_TRANS_READY)) {
				//-----------------------------------------------------------------------------
				// [XBEE_TRANS_READY]
				//-----------------------------------------------------------------------------
				if (XBeePacketHeader.cbExtra == 2) {	// we need to read in a new DL for the transmitter...
					byte abNewDL[2];
					cbRead = ReadFromXBee((byte*)&abNewDL, (byte)sizeof(abNewDL), (ulong)10000, (word)-1);
					if (cbRead == sizeof(abNewDL)) {
						bcsIn = XBeePacketHeader.bSeqNum + XBeePacketHeader.cbExtra + abNewDL[0] + abNewDL[1];

						if (XBeePacketHeader.bChksum == bcsIn) {
							// Currently BAP sends 16 bit numbers MSB/LSB so lets build our actual wNewDL
							wNewDL = ((word)abNewDL[0] << 8) + (word)abNewDL[1];
							g_diystate.fTransReadyRecvd = true;		// OK we have received a packet saying transmitter is ready.	
							SetXBeeDL(wNewDL);
							pause(20);					// give time to receive the oks...
							ClearXBeeInputBuffer();		// get rid of all of the OKs...

							// And tell the remote to go into New data only mode.
							SendXbeeNewDataOnlyPacket(1);
							g_diystate.ulLastPacket = GetCurrentTime();
							g_diystate.fReqDataPacketSent = 0;							// make sure we don't think we have any outstanding requests
							_fNewPacketAvail = true;									// and try to get the first packet of data
#ifdef DEBUG
							if (g_fDebugOutput) {
								DBGPrintf("XBee_Trans_READY: %x\n", wNewDL);
							}
#endif //DEBUG
						}
					}
				}
			} else if ((XBeePacketHeader.bPacketType == XBEE_TRANS_NOTREADY) && (XBeePacketHeader.bChksum == 0)  
				&& (XBeePacketHeader.bSeqNum == 0) 	&& (XBeePacketHeader.cbExtra == 0)) {
				//-----------------------------------------------------------------------------
				// [XBEE_TRANS_NOTREADY]
				//-----------------------------------------------------------------------------
				g_diystate.fTransReadyRecvd = 0;			// Ok not valid anymore...
#ifdef DEBUG
				if (g_fDebugOutput) {
					DBGPrintf("XBee_Trans_NOTREADY\r\n");
				}
#endif //DEBUG
			}	

			//-----------------------------------------------------------------------------
			// [XBEE_TRANS_NEW]
			//-----------------------------------------------------------------------------
			else if ((XBeePacketHeader.bPacketType == XBEE_TRANS_NEW) && (XBeePacketHeader.bChksum == 0) && 
					(XBeePacketHeader.bSeqNum == 0) && (XBeePacketHeader.cbExtra == 0)) {
				// The remote has told us it has new data available for us, so we should now ask for it!
				//-----------------------------------------------------------------------------
				// [XBEE_TRANS_NEW]
				//-----------------------------------------------------------------------------
				_fNewPacketAvail = true;
			} else {
				//-----------------------------------------------------------------------------
				// [UNKNOWN PACKET]
				//-----------------------------------------------------------------------------
#ifdef DEBUG
				if (g_fDebugOutput) {
					DBGPrintf("Unknown Packet: %x %x %x %x\n", XBeePacketHeader.bPacketType, XBeePacketHeader.bChksum,
							XBeePacketHeader.bSeqNum, XBeePacketHeader.cbExtra);
				}
#endif //DEBUG

				ClearXBeeInputBuffer();
			}
		}
	}

//-----------------------------------------------------------------------------
// [See if we need to request data from the other side]
//-----------------------------------------------------------------------------
	// Only send when we know the transmitter is ready.  Also if we are in the New data only mode don't ask for data unless we have been told there
	// is new data. We relax this a little and be sure to ask for data every so often as to make sure the remote is still working...
	// 
	if (g_diystate.fTransReadyRecvd) {
		ulCurrentTime = GetCurrentTime();

		// Time in MS since last packet
		ulTimeDiffMS = ((ulCurrentTime-g_diystate.ulLastPacket)*WTIMERTICSPERMSMUL) / WTIMERTICSPERMSDIV;

		// See if we exceeded a global timeout.  If so let caller know so they can stop themself if necessary...
		if (ulTimeDiffMS > CXBEETIMEOUTRECVMS) {
#ifdef DEBUG
			if (g_fDebugOutput) {
				if (!s_fDisplayedTimeout) {
					DBGPrintf("XBEE Timeout\n"); 
					s_fDisplayedTimeout = true;
				}
			}				
#endif //DEBUG
			g_diystate.fPacketValid = 0;
			g_diystate.fPacketForced = true;
			return false;
		}

		// see if we have an outstanding request out and if it timed out...
		if (g_diystate.fReqDataPacketSent) {
			if ((((ulCurrentTime-g_diystate.ulLastRequest) * WTIMERTICSPERMSMUL) / WTIMERTICSPERMSDIV) > CXBEEPACKETTIMEOUTMS) {
				// packet request timed out, force a new attempt.
				_fNewPacketAvail = true;		// make sure it requests a new one	
			}
		}
		
		// Next see if it has been too long since we received a packet.  Ask to make sure they are there...
		if (!_fNewPacketAvail && (ulTimeDiffMS > CXBEEFORCEREQMS)) {
			_fNewPacketAvail = true;
			g_diystate.fReqDataForced = true;		// remember that this request was forced!
		}

		if (!g_diystate.fSendOnlyNewMode || (g_diystate.fSendOnlyNewMode && _fNewPacketAvail)) {
			// Now send out a prompt request to the transmitter:
			g_diystate.bPacketNum++;		// rolling counter 

			SendXBeePacket (XBEE_RECV_REQ_DATA, g_diystate.bPacketNum,  0, 0);	// Request data Prompt (CmdType, ChkSum, Packet Number, CB extra data)
			g_diystate.fReqDataPacketSent = true; 									// yes we have already sent one.
			g_diystate.ulLastRequest = ulCurrentTime;						// remember when we sent this...
		}
		g_diystate.fPacketValid = _fPacketValidPrev;	// Say the data is in the same state as the previous call...
	}

	return g_diystate.fPacketValid;
}

#ifdef __cplusplus
}
#endif
